import asyncio
import logging
from collections import defaultdict
from typing import Dict
from uuid import UUID

from fastapi import FastAPI
from smartcard.CardConnection import CardConnection
from smartcard.CardMonitoring import CardMonitor
from smartcard.Exceptions import SmartcardException
from smartcard.ReaderMonitoring import ReaderMonitor
from starlette.websockets import WebSocket, WebSocketDisconnect

from pyscard_json_rpc import connections
from pyscard_json_rpc.json_rpc import format_error, get_message_type, format_response
from pyscard_json_rpc.methods import METHOD_HANDLERS
from pyscard_json_rpc.observers import PrintReaderObserver, PrintCardObserver

logger = logging.getLogger(__name__)


def environment_setup():
    pass


def logging_setup():
    # logging.basicConfig(level=logging.INFO)
    # logger.setLevel(logging.INFO)
    pass


app = FastAPI(on_startup=[environment_setup, logging_setup], on_shutdown=[])

### tachocard app -> smartcard service
# smartcard.get_readers
# smartcard.get_cards
# smartcard.connect
# smartcard.get_atr
# smartcard.get_apdu
# smartcard.disconnect

### smartcard service -> tachocard app
# observer.reader_added
# observer.reader_removed
# observer.card_added
# observer.card_removed


@app.websocket("/ws/pyscard-json-rpc")
async def websocket_handler(websocket: WebSocket):

    await websocket.accept()
    # define monitors/observers
    loop = asyncio.get_event_loop()
    readermonitor = ReaderMonitor()
    readerobserver = PrintReaderObserver(websocket=websocket, loop=loop)
    cardmonitor = CardMonitor()
    cardobserver = PrintCardObserver(websocket=websocket, loop=loop)

    try:
        # add observers
        readermonitor.addObserver(readerobserver)
        cardmonitor.addObserver(cardobserver)
        while True:
            message = await websocket.receive_json()
            message_type = get_message_type(message)

            if message_type == "request":
                if message["method"] in METHOD_HANDLERS:
                    try:
                        result = await METHOD_HANDLERS[message["method"]](
                            websocket=websocket, **message.get("params", {})
                        )
                        response = format_response(result=result, request_id=message.get("id"))
                    except (KeyError, StopIteration) as exc:
                        logger.exception(f"caught exception during handling {message['method']}", exc_info=exc)
                        response = format_error(code=-32602, request_id=message.get("id"))
                    except SmartcardException as exc:
                        logger.exception(f"caught exception during handling {message['method']}", exc_info=exc)
                        response = format_error(code=-32000, message=exc.args[0], request_id=message.get("id"))
                else:
                    response = format_error(code=-32601, request_id=message.get("id"))

                await websocket.send_json(response)
            elif message_type in {"success", "error"}:
                pass

    except ValueError as exc:
        logger.exception(exc)
        await websocket.close(code=1003)
    except WebSocketDisconnect:
        pass
    finally:
        connection_id: UUID
        connection: CardConnection
        for connection_id, connection in connections.card_connections[websocket.scope["client"]].items():
            logger.error(f"connection {connection_id} was left open")
            connection.disconnect()

        # remove observers
        readermonitor.deleteObserver(readerobserver)
        cardmonitor.deleteObserver(cardobserver)



from typing import Optional, Dict

ERROR_CODES = {
    # our codes:
    423: "Locked",
    428: "Precondition Required",
    # JSON RPC Prefedined codes
    -32700: "Parse error",
    # Invalid JSON was received by the server. An error occurred on the server while parsing the JSON text.
    -32600: "Invalid Request",  # The JSON sent is not a valid Request object.
    -32601: "Method not found",  # The method does not exist / is not available.
    -32602: "Invalid params",  # Invalid method parameter(s).
    -32603: "Internal error",  # Internal JSON-RPC error.
    # -32000 to -32099:	"Server error"  # Reserved for implementation-defined server-errors.
}


def get_message_type(message: Dict) -> str:
    if message.get("jsonrpc") == "2.0":
        if "method" in message:
            return "request"
        elif "result" in message:
            return "success"
        elif "error" in message:
            return "error"
    raise ValueError("Message does not conform to JSON-RPC 2.0 standard")


def format_request(method: str, request_id: Optional[str] = None, params: Optional[Dict] = None) -> Dict:
    data = {
        "jsonrpc": "2.0",
        "method": method,
    }
    if request_id is not None:
        data["id"] = request_id
    if params is not None:
        data["params"] = params
    return data


def format_error(
    code: int, message: Optional[str] = None, data: Optional[str] = None, request_id: Optional[str] = None
) -> Dict:
    error = {"code": code, "message": message or ERROR_CODES.get(code) or "Unknown error"}
    if data is not None:
        error["data"] = data
    data = {
        "jsonrpc": "2.0",
        "error": error,
    }
    if request_id is not None:
        data["id"] = request_id
    return data


def format_response(result, request_id: Optional[str]) -> Dict:
    data = {
        "jsonrpc": "2.0",
        "result": result,
    }
    if request_id is not None:
        data["id"] = request_id
    return data



from collections import defaultdict
from typing import Dict, Tuple
from uuid import UUID

from smartcard.CardConnection import CardConnection

card_connections: Dict[Tuple[str, int], Dict[UUID, CardConnection]] = defaultdict(dict)



"""Console script for pyscard_json_rpc."""
import argparse
import sys
import uvicorn

from pyscard_json_rpc.app import app


def main():
    """Console script for pyscard_json_rpc."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-H", "--host", help="host ip", default="0.0.0.0")
    parser.add_argument("-p", "--port", help="port of the web server", type=int, default=8040)
    parser.add_argument("_", nargs="*")
    args = parser.parse_args()

    uvicorn.run(app, host=args.host, port=args.port, log_level="debug")


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import binascii
import struct
import uuid
from uuid import UUID

import smartcard.System
from smartcard.CardConnection import CardConnection
from smartcard.reader.Reader import Reader

from pyscard_json_rpc import connections


async def smartcard_get_readers(websocket, **params):
    return {"readers": [reader.name for reader in smartcard.System.readers()]}


async def smartcard_connect(websocket, **params):
    readers = smartcard.System.readers()
    reader: Reader = next(reader for reader in readers if reader.name == params.get("reader"))

    connection_id: UUID = uuid.uuid4()
    connection: CardConnection = reader.createConnection()
    connections.card_connections[websocket.scope["client"]][connection_id] = connection

    connection.connect(
        protocol=params.get("protocol"),
        mode=params.get("mode"),
        disposition=params.get("disposition"),
    )

    return {
        "connection_id": str(connection_id),
    }


async def connection_get_atr(websocket, **params):
    connection_id = UUID(params["connection_id"])
    connection: CardConnection = connections.card_connections[websocket.scope["client"]][connection_id]

    atr = connection.getATR()

    return {
        "atr": binascii.hexlify(bytes(atr)).decode("utf-8"),
    }


async def connection_transmit(websocket, **params):
    connection_id = UUID(params["connection_id"])
    connection: CardConnection = connections.card_connections[websocket.scope["client"]][connection_id]

    apdu_request = binascii.unhexlify(params["apdu"])
    data, sw1, sw2 = connection.transmit(
        bytes=[int(x) for x in apdu_request or []],
        protocol=params.get("protocol"),
    )

    return {
        "data": binascii.hexlify(bytes(data)).decode("utf-8"),
        "sw1": binascii.hexlify(struct.pack("<B", sw1)).decode("utf-8"),
        "sw2": binascii.hexlify(struct.pack("<B", sw2)).decode("utf-8"),
    }


async def connection_disconnect(websocket, **params):
    connection_id = UUID(params["connection_id"])
    connection: CardConnection = connections.card_connections[websocket.scope["client"]][connection_id]
    connection.disconnect()

    del connections.card_connections[websocket.scope["client"]][connection_id]

    return {}


METHOD_HANDLERS = {
    "smartcard.get_readers": smartcard_get_readers,
    "smartcard.connect": smartcard_connect,
    "connection.get_atr": connection_get_atr,
    "connection.transmit": connection_transmit,
    "connection.disconnect": connection_disconnect,
}



#!/usr/bin/env python

"""The setup script."""
from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()


def get_requirements(filename):
    """
    parse requirements.txt, ignore links, exclude comments
    """
    requirements = []
    for line in open(filename).readlines():
        # skip to next iteration if comment or empty line
        if line.startswith("#") or line == "" or line.startswith("http") or line.startswith("git"):
            continue
        # add line to requirements
        requirements.append(line)
    return requirements


setup(
    author="Namespace OÜ",
    author_email="info@namespace.ee",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="JSON RPC bindings for pyscard.",
    entry_points={
        "console_scripts": [
            "pyscard_json_rpc=pyscard_json_rpc.cli:main",
        ],
    },
    install_requires=get_requirements("requirements.txt"),
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="pyscard-json-rpc",
    name="pyscard-json-rpc",
    packages=find_packages(include=["pyscard_json_rpc", "pyscard_json_rpc.*"]),
    test_suite="tests",
    tests_require=get_requirements("requirements_dev.txt"),
    url="https://github.com/namespace-ee/pyscard-json-rpc",
    version="0.2.0",
    zip_safe=False,
)



from __future__ import print_function

import asyncio
import uuid

from smartcard.ATR import ATR
from smartcard.util import toHexString, PACK
from smartcard.CardMonitoring import CardObserver
from smartcard.ReaderMonitoring import ReaderObserver

from pyscard_json_rpc.json_rpc import format_request

PROTOCOLS_MAPPING = {
    "T=0": 0x00000001,  # T0_protocol
    "T=1": 0x00000002,  # T1_protocol
    "T=15": 0x00000008,  # T15_protocol
}


class PrintReaderObserver(ReaderObserver):
    """A simple reader observer that is notified
    when readers are added/removed from the system and
    prints the list of readers
    """

    websocket = None
    loop = None

    def __init__(self, websocket, loop):
        super().__init__()
        self.websocket = websocket
        self.loop = loop

    def update(self, observable, actions):
        addedreaders, removedreaders = actions
        for reader in addedreaders or []:
            response = format_request(
                method="observer.reader_added",
                request_id=str(uuid.uuid4()),
                params={"reader": reader.name},
            )
            asyncio.run_coroutine_threadsafe(self.websocket.send_json(response), self.loop)

        for reader in removedreaders or []:
            response = format_request(
                method="observer.reader_removed",
                request_id=str(uuid.uuid4()),
                params={"reader": reader.name},
            )
            asyncio.run_coroutine_threadsafe(self.websocket.send_json(response), self.loop)


class PrintCardObserver(CardObserver):
    """A simple card observer that is notified
    when cards are inserted/removed from the system and
    prints the list of cards
    """

    websocket = None
    loop = None

    def __init__(self, websocket, loop):
        super().__init__()
        self.websocket = websocket
        self.loop = loop

    def update(self, observable, actions):
        addedcards, removedcards = actions
        for card in addedcards:
            atr = ATR(card.atr)
            response = format_request(
                method="observer.card_added",
                request_id=str(uuid.uuid4()),
                params={
                    "reader": str(card.reader),
                    "atr": toHexString(card.atr, PACK),
                    "supported_protocols": [
                        PROTOCOLS_MAPPING[k]
                        for k, v in atr.getSupportedProtocols().items()
                        if k in PROTOCOLS_MAPPING and v is True
                    ],
                },
            )
            asyncio.run_coroutine_threadsafe(self.websocket.send_json(response), self.loop)

        for card in removedcards:
            response = format_request(
                method="observer.card_removed",
                request_id=str(uuid.uuid4()),
                params={"reader": str(card.reader), "atr": toHexString(card.atr, PACK)},
            )
            asyncio.run_coroutine_threadsafe(self.websocket.send_json(response), self.loop)
